﻿using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;

using GHIElectronics.NETMF.Net;
using GHIElectronics.NETMF.Net.NetworkInformation;
using GHIElectronics.NETMF.Net.Sockets;
using GHIElectronics.NETMF.FEZ;
using GHIElectronics.NETMF.Hardware;

using System;
using System.Text;
using System.Threading;



namespace MFConsoleApplication1 {
    public class NetworkBuzzer {
        // Global declaration of variable lcd (to access the lcd object).
        static FEZ_Components.FEZTouch lcd;

        /// <summary>
        /// The Main() procedure will handle the initialisation, the gaining of the IP address
        /// and the running of the program.
        /// </summary>
        public static void Main()
        {
            /********************************************************************************************************
             *  Default Network Settings
             *  
             *  Here you can set the properties of your network connection.
             *  Set USE_DHCP to true if you want to obtain an IP address from the DHCP server.
             *  Otherwise set it to false: then given network settings will be applied.
             ********************************************************************************************************/
            bool USE_DHCP = true;     // >> set to true or false.
            String HOST = "panda";                                // hostname of the board
            byte[] IP = { 192, 168, 2, 111 };                     // only applicabe when DHCP is off
            byte[] SUBNET = { 255, 255, 255, 0 };                 // only applicabe when DHCP is off
            byte[] GATEWAY = { 192, 168, 2, 1 };                  // only applicabe when DHCP is off
            byte[] MAC = { 0x00, 0x26, 0x1C, 0x7B, 0x29, 0xE8 };  // MAC address of the network interface
            int UDP_PORT = 2000;                                   // UDP port of the server


            // Retrieve the clock from the Real Time Clock module.
            Utility.SetLocalTime(RealTimeClock.GetTime());

            // Define everything there is needed to get the LCD working.
            FEZ_Components.FEZTouch.LCDConfiguration lcdConfig = new FEZ_Components.FEZTouch.LCDConfiguration(
                FEZ_Pin.Digital.Di28,
                FEZ_Pin.Digital.Di20,
                FEZ_Pin.Digital.Di22,
                FEZ_Pin.Digital.Di23,
                new FEZ_Pin.Digital[8] { FEZ_Pin.Digital.Di51, 
                                         FEZ_Pin.Digital.Di50, 
                                         FEZ_Pin.Digital.Di49, 
                                         FEZ_Pin.Digital.Di48, 
                                         FEZ_Pin.Digital.Di47, 
                                         FEZ_Pin.Digital.Di46, 
                                         FEZ_Pin.Digital.Di45, 
                                         FEZ_Pin.Digital.Di44 },
                FEZ_Pin.Digital.Di24,
                FEZ_Pin.Digital.Di26

                );
            FEZ_Components.FEZTouch.TouchConfiguration touchConfig = new FEZ_Components.FEZTouch.TouchConfiguration(
                SPI.SPI_module.SPI2, FEZ_Pin.Digital.Di25, FEZ_Pin.Digital.Di34
                );
            lcd = new FEZ_Components.FEZTouch(lcdConfig, touchConfig);


            // Show the initial screen
            showBootMessage("Starting remote buzzer ..."); 
  

            // Activate the network Interface.
            WIZnet_W5100.Enable(SPI.SPI_module.SPI1, (Cpu.Pin)FEZ_Pin.Digital.Di10,(Cpu.Pin)FEZ_Pin.Digital.Di7,  true);


            // The following section depends on desired setting: Use DHCP or not?
            if (USE_DHCP==true)
            {   // DHCP is being used. Try to receive an IP address from the network.                                                        
                try
                {
                    Dhcp.EnableDhcp(MAC, HOST);
                }
                catch (Exception e)
                {
                    Debug.Print("An error occured!" + e.Message);
                }
            }
            else
            {   // A static IP is being used. 
                NetworkInterface.EnableStaticIP(IP, SUBNET, GATEWAY, MAC);
            } // (USE_DHCP)

          
            // Show the first screen with all information.
            String currentIP = new IPAddress(NetworkInterface.IPAddress).ToString();
            showScreen(currentIP, UDP_PORT.ToString(),  "0", "0.0.0.0");    // frequency and requester are unknown yet.


            // Create a socket at the desired port.
            Socket serversocket = new Socket(AddressFamily.InterNetwork, SocketType.Dgram, ProtocolType.Udp);
            EndPoint remoteEndPoint = new IPEndPoint(IPAddress.Any, UDP_PORT);
            serversocket.Bind(remoteEndPoint);


            // Enable a PWM-output to control the buzzer.
            PWM buzzer = new PWM((PWM.Pin)FEZ_Pin.PWM.Di5);

          
            // PROGRAM LOOP
            // This loop will be cycled to poll the server socket
            while (true) {
                if (serversocket.Poll(-1, SelectMode.SelectRead))
                {
                    // Retrieve the data and put it into a String called 'result'.
                    byte[] inBuf = new byte[serversocket.Available];
                    int count = serversocket.ReceiveFrom(inBuf, ref remoteEndPoint);
                    IPEndPoint remoteInfo = (IPEndPoint)remoteEndPoint;                   
                    String result = new String(Encoding.UTF8.GetChars(inBuf));
                    
                    // Split the String into an array of words.
                    String[] items = result.Split(' ');

                    // Protocol check: does the datagram have the right number of fields?
                    if (items.Length == 2)
                    { // Received datagram has the right appearence: 2 text fields.

                        // Protocol check: does the datagram starts with the word FREQ?
                        if (items[0].ToUpper().Trim() == "FREQ")
                        {   // Protocol first word OK
                            try
                            {   // Try to convert the frequency into an Integer.
                                int frequency = int.Parse(items[1]);
                                if (frequency >= 0 && frequency <= 15000)
                                {   // Frequency is within range
                                    buzzer.Set(frequency, 50);
                                }
                                else
                                {   // Frequency is outside range (0 ... 15 kHz)
                                    Debug.Print("Received datagram contains an invalid frequence (outside range)");
                                } // if (frequency >= 0 && frequency <= 15000)                             
                            }
                            catch
                            {   // Parsing to Integer failed: invalid integer value
                                Debug.Print("Received datagram contains an invalid frequence (no integer value)");
                            } // try

                            // Show the information on the screen
                            refreshScreen(currentIP, UDP_PORT.ToString(), items[1], remoteInfo.Address.ToString());
                        }
                        else
                        {   // Error: datagram doesn't start with FREQ
                            Debug.Print("Received datagram doesn't start with FREQ");
                        } // if (items[0].ToUpper().Trim() == "FREQ")
                    }
                    else
                    {   // Received datagram has an invalid number of fields
                        Debug.Print("Received datagram has an invalid number of fields");
                    }
                } // if (serversocket.Poll(-1, SelectMode.SelectRead))
            }   // while (true)
        }



        /// <summary>
        /// ShowBootMessage will show the first informations on a black screen. It will only
        /// appear for a short amount of time.
        /// </summary>
        /// <param name="message">The message to show</param>
        private static void showBootMessage(String message)
        {
            lcd.FillRectangle(0, 0, 240, 320, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString(message, 0, 0, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
        }


        /// <summary>
        /// showScreen will show the main screen with all informations. The parameters passed
        /// to this function will be showed on the screen.
        /// </summary>
        /// <param name="serverIP">The current IP address of the server</param>
        /// <param name="serverPort">The port of the server</param>
        /// <param name="frequency">The current frequency that is running</param>
        /// <param name="requester">The IP address of the requester</param>
        private static void showScreen(String serverIP, String serverPort, String frequency, String requester)
        {
            lcd.FillRectangle(0, 0, 240, 320, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("REMOTE BUZZER APP", 0, 0, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("-----------------", 0, 15, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Server:" + serverIP, 0, 30, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Port:" + serverPort, 0, 45, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("OUTPUT STATUS:", 0, 75, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Frequency:" + frequency + " Hz", 0, 90, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Requester:" + requester, 0, 105, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
        }


        /// <summary>
        /// refreshScreen will show the main screen with all informations. Unlike the function showScreen it will
        /// only show the updated informations about frequency and requester.
        /// </summary>
        /// <param name="serverIP">The current IP address of the server</param>
        /// <param name="serverPort">The port of the server</param>
        /// <param name="frequency">The current frequency that is running</param>
        /// <param name="requester">The IP address of the requester</param>
        private static void refreshScreen(String serverIP, String serverPort, String frequency, String requester)
        {
            lcd.DrawString("Frequency:" + frequency + " Hz     ", 0, 90, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Requester:" + requester + "     ", 0, 105, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);
            lcd.DrawString("Time:" + DateTime.UtcNow.ToString() + "     ", 0, 120, FEZ_Components.FEZTouch.Color.Yellow, FEZ_Components.FEZTouch.Color.Blue);          
        }
    }
}
